/*

	wmsysmon -- system monitoring dockapp

	Copyright (C) 1998-1999 Dave Clark - clarkd@skynet.ca
        Copyright (C) 2000 Vito Caputo - swivel@gnugeneration.com

        wmsysmon is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        wmsysmon is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>

#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "wmgeneral.h"

#ifdef HI_INTS
#include "wmsysmon-master-alpha.xpm"
#else
#include "wmsysmon-master-i386.xpm"
#endif


char wmsysmon_mask_bits[64*64];
int  wmsysmon_mask_width = 64;
int  wmsysmon_mask_height = 64;

long start_time=0;
long start_uptime=0;

long io_max;
long io_max_diff;

#define WMSYSMON_VERSION "0.5.1"

#define CHAR_WIDTH 5
#define CHAR_HEIGHT 7

#define BCHAR_WIDTH 6
#define BCHAR_HEIGHT 9

#define BOFFX   (93)
#define BOFFY   (66)
#define BREDX   (81)
#define BREDY   (66)
#define BGREENX (87)
#define BGREENY (66)

#define LITEW   (4)
#define LITEH   (4)

#define B_OFF   (0)
#define B_RED   (1)
#define B_GREEN (2)


#ifdef HI_INTS
long last_ints[24];
long ints[24];
#else
long last_ints[16];
long ints[16];
#endif

long last_pageins=0;
long last_pageouts=0;

long last_swapins=0;
long last_swapouts=0;

FILE *statfp;
FILE *memfp;

int update_rate=300000;

extern	char **environ;

char	*ProgName;

char uconfig_file[256];

time_t		curtime;
time_t		prevtime;

void usage(void);
void printversion(void);
void BlitString(char *name, int x, int y);
void BlitNum(int num, int x, int y);
void wmsysmon_routine(int, char **);
int ReadConfigInt(FILE *fp, char *setting, int *value);
int ReadConfigString(FILE *fp, char *setting, char *value);
int Read_Config_File( char *filename );
void DrawBar(int sx, int sy, int w, int h, float percent, int dx, int dy);
inline void DrawLite(int state, int dx, int dy);
inline void DrawUptime(void);
inline void DrawStuff( void );
inline void DrawMem( void );

int main(int argc, char *argv[]) {
	int		i;
	uconfig_file[0] = 0;
    
	/* Parse Command Line */

	ProgName = argv[0];
	if (strlen(ProgName) >= 5)
		ProgName += (strlen(ProgName) - 5);

	for (i=1; i<argc; i++) {
		char *arg = argv[i];

		if (*arg=='-') {
			switch (arg[1]) {
			case 'd' :
				if (strcmp(arg+1, "display")) {
					usage();
					exit(1);
				}
				break;
			case 'g' :
				if (strcmp(arg+1, "geometry")) {
					usage();
					exit(1);
				}
				break;
			case 'v' :
				printversion();
				exit(0);
				break;
			case 'r':
				if (argc > (i+1)) {
					update_rate = (atoi(argv[i+1]) * 1000);
					i++;
				}
				break;
			case 'c' :
				if (argc > (i+1)) {
					strcpy(uconfig_file, argv[i+1]);
					i++;
				}
				break;
			default:
				usage();
				exit(0);
				break;
			}
		}
	}

	wmsysmon_routine(argc, argv);

	return 0;
}


void wmsysmon_routine(int argc, char **argv)
{
	int		i;
	XEvent		Event;
	int		but_stat = -1;
	FILE		*fp;
	struct timeval	tv={0,0};
	struct timeval	last={0,0};
    

	createXBMfromXPM(wmsysmon_mask_bits, wmsysmon_master_xpm, wmsysmon_mask_width, wmsysmon_mask_height);
    
	openXwindow(argc, argv, wmsysmon_master_xpm, wmsysmon_mask_bits, wmsysmon_mask_width, wmsysmon_mask_height);

    /* init ints */
	bzero(&last_ints, sizeof(last_ints));
	bzero(&ints, sizeof(ints));

    
    /* init uptime */
	fp = fopen("/proc/uptime", "r");
	if (fp) {
		fscanf(fp, "%ld", &start_time);
		fclose(fp);
		start_uptime = time(0);
	}

	RedrawWindow();


	statfp = fopen("/proc/stat", "r");
	memfp = fopen("/proc/meminfo", "r");

	while (1) {
        
		curtime = time(0);

		if(1) {

			memcpy(&last, &tv, sizeof(tv));
            
			// Update display

			DrawUptime();

			DrawStuff();

			DrawMem();

			RedrawWindow();
		}
        
		/* X Events */
		while (XPending(display)) {
			XNextEvent(display, &Event);
			switch (Event.type) {
			case Expose:
				RedrawWindow();
				break;
			case DestroyNotify:
				XCloseDisplay(display);
				exit(0);
				break;
			case ButtonPress:
				i = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);

				but_stat = i;
				break;
			case ButtonRelease:
				i = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);

				if (but_stat == i && but_stat >= 0) {
					switch (but_stat) {
					case 0 :
						break;
					case 1 :
						break;
					case 2:
						break;
					case 3:
						break;
					case 4:
						break;

					}
				}
				but_stat = -1;
				break;
			}
		}

		usleep(update_rate);
	}
}


inline void DrawBar(int sx, int sy, int w, int h, float percent, int dx, int dy)
{
	static int tx;

#ifdef MONDEBUG
	puts("DrawBar()");
#endif

	tx = (float)((float)w * ((float)percent / (float)100.0));

    
	copyXPMArea(sx, sy, tx, h, dx, dy);

	copyXPMArea(sx, sy+h+1, w-tx, h, dx+tx, dy);
}


inline void DrawLite(int state, int dx, int dy)
{

#ifdef MONDEBUG
	puts("DrawLite()");
#endif

	switch(state) {
	case B_RED:
		copyXPMArea(BREDX, BREDY, LITEW, LITEH, dx, dy);
		break;
	case B_GREEN:
		copyXPMArea(BGREENX, BGREENY, LITEW, LITEH, dx, dy);
		break;
	default:
	case B_OFF:
		copyXPMArea(BOFFX, BOFFY, LITEW, LITEH, dx, dy);
		break;
	}
}


inline void DrawUptime(void)
{
	static int first = 1;
	static int old_days = 0, old_hours = 0, old_minutes = 0;
	static long uptime;
    	static char buf[32];
    	static int i;
    
	uptime = curtime - start_uptime + start_time;

	uptime /=60;
	/* blit minutes */
	i = uptime % 60;
	uptime /=60;

	if(old_minutes != i || first) {
		old_minutes = i;
		sprintf(buf, "%02i", i);
#ifdef HI_INTS
		BlitString(buf, 45, 29);
#else
		BlitString(buf, 45, 37);
#endif
	}
    
	/* blit hours */
	i = uptime % 24;
	uptime /=24;

	if(old_hours != i || first) {
		old_hours = i;
		sprintf(buf, "%02i", i);
#ifdef HI_INTS
		BlitString(buf, 29, 29);
#else
		BlitString(buf, 29, 37);
#endif
	}

    
	/* blit days */
	i = uptime;
	if(old_days != i || first) {
		old_days = i;
		sprintf(buf, "%03i", i);
#ifdef HI_INTS
		BlitString(buf, 6, 29);
#else
		BlitString(buf, 6, 37);
#endif
	}

	first = 0;
}


inline void DrawStuff( void )
{
	static int io_last = 0, first = 1;
#ifdef HI_INTS
	static int int_lites[24] =
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		/* to keep track of on/off status */
#else
    	static int int_lites[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#endif
	static int pagein_lite = 0, pageout_lite = 0, swapin_lite = 0, swapout_lite = 0;
    	static char buf[1024];
    	static char *tok;
	char seps[]={" "};
	static int i;

	static long io;
	static long iodiff;
	static int iopercent;

	static long pageins;
	static long pageouts;
	static long swapins;
	static long swapouts;

	io = iodiff = iopercent = pageins = pageouts = swapins = swapouts = 0;

	statfp = freopen("/proc/stat", "r", statfp);

	if(!statfp) return;

	/* gather data... */
	while( fgets(buf, 127, statfp) ) {
		if (strstr(buf, "disk_rio")) {
			tok = strtok(buf, seps);
			for(i=0; i<4; i++) {
				io += atoi(tok);
				tok = strtok(NULL, seps);
			}
		}

		if(strstr(buf, "disk_wio")) {
			tok = strtok(buf, seps);
			for(i=0; i<4; i++) {
				io += atoi(tok);
				tok = strtok(NULL, seps);
			}
		}

		if (strstr(buf, "intr")) {

#ifdef HI_INTS
			sscanf(buf, "intr %*d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
			&ints[0], &ints[1], &ints[2], &ints[3], &ints[4],
			&ints[5], &ints[6], &ints[7], &ints[8], &ints[9],
			&ints[10], &ints[11], &ints[12], &ints[13], &ints[14],
			&ints[15], &ints[16], &ints[17], &ints[18], &ints[19],
			&ints[20], &ints[21], &ints[22], &ints[23]);
#else
			sscanf(buf, "intr %*d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
			&ints[0], &ints[1], &ints[2], &ints[3], &ints[4],
			&ints[5], &ints[6], &ints[7], &ints[8], &ints[9],
			&ints[10], &ints[11], &ints[12], &ints[13], &ints[14],
			&ints[15]);
#endif
		}

		if (strstr(buf, "page")) {
			sscanf(buf, "page %ld %ld", &pageins, &pageouts);
		}

		if (strstr(buf, "swap")) {
			sscanf(buf, "swap %ld %ld", &swapins, &swapouts);
		}
	}


	/* ------------------ IO bar ------------------ */
	if(io_max == 0)
		io_max = io;
    
	if(io > io_max)
		iodiff = abs(io_max - io);
	else
		iodiff=0;

	io_max = io;

    
	if(io_max_diff !=0)
		iopercent = ((float) iodiff / (float) io_max_diff) * 100.0;
	else
		iopercent=0;

	if(iodiff > io_max_diff)
		io_max_diff=iodiff;

	if (iopercent >100)
		iopercent =100;
    
/*
	printf("io: %i\n", io);
	printf("io_max: %i\n", io_max);
	printf("iodiff: %i\n", iodiff);
	printf("iopercent: %i\n", iopercent);
*/

	if(iopercent != io_last || first) {
		io_last = iopercent;
#ifdef HI_INTS
		DrawBar(67, 36, 58, 4, iopercent, 3, 20);
#else
		DrawBar(67, 36, 58, 6, iopercent, 3, 26);
#endif
	}

    /* top 8 ints */
	for (i=0; i<8; i++) {
		if ( ints[i] > last_ints[i] && !int_lites[i]) {
			int_lites[i] = 1;
#ifdef HI_INTS
			DrawLite(B_GREEN, 4 + (i*LITEW)+i, 43);
#else
			DrawLite(B_GREEN, 4 + (i*LITEW)+i, 51);
#endif
		} else if(ints[i] == last_ints[i] && int_lites[i]) {
			int_lites[i] = 0;
#ifdef HI_INTS
			DrawLite(B_OFF, 4 + (i*LITEW)+i, 43);
#else
			DrawLite(B_OFF, 4 + (i*LITEW)+i, 51);
#endif
		}
	}
    /* middle/bottom 8 */
	for (i=8; i<16; i++) {
		if ( ints[i] > last_ints[i] && !int_lites[i]) {
			int_lites[i] = 1;
#ifdef HI_INTS
			DrawLite(B_GREEN, 4 + ((i-8)*LITEW)+(i-8), 48);
#else
			DrawLite(B_GREEN, 4 + ((i-8)*LITEW)+(i-8), 56);
#endif
		} else if(ints[i] == last_ints[i] && int_lites[i]) {
			int_lites[i] = 0;
#ifdef HI_INTS
			DrawLite(B_OFF, 4 + ((i-8)*LITEW)+(i-8), 48);
#else
			DrawLite(B_OFF, 4 + ((i-8)*LITEW)+(i-8), 56);
#endif
		}
	}

#ifdef HI_INTS
    /* bottom 8 on alpha/smp x86 */
	for (i=16;i<24;i++) {
		if (ints[i] > last_ints[i] && !int_lites[i] ) {
			int_lites[i] = 1;
			DrawLite(B_GREEN, 4 + ((i-16)*LITEW)+(i-16),53);
		} else if(ints[i] == last_ints[i] && int_lites[i]) {
			int_lites[i] = 0;
			DrawLite(B_OFF,4+((i-16)*LITEW)+(i-16),53);
		}
	}
#endif

	memcpy(&last_ints, &ints, sizeof(ints));

	/* page in / out */

	if (pageins > last_pageins && !pagein_lite) {
		pagein_lite = 1;
#ifdef HI_INTS
		DrawLite(B_RED, 51, 43);
#else
		DrawLite(B_RED, 51, 51);
#endif
	} else if(pagein_lite) {
		pagein_lite = 0;
#ifdef HI_INTS
		DrawLite(B_OFF, 51, 43);
#else
		DrawLite(B_OFF, 51, 51);
#endif
	}

	if (pageouts > last_pageouts && !pageout_lite) {
		pageout_lite = 1;
#ifdef HI_INTS
		DrawLite(B_RED, 56, 43);
#else
		DrawLite(B_RED, 56, 51);
#endif
	} else if(pageout_lite) {
		pageout_lite = 0;
#ifdef HI_INTS
		DrawLite(B_OFF, 56, 43);
#else
		DrawLite(B_OFF, 56, 51);
#endif
	}

	last_pageins = pageins;
	last_pageouts = pageouts;

	/* swap in/out */

	if(swapins > last_swapins && !swapin_lite) {
		swapin_lite = 1;
#ifdef HI_INTS
		DrawLite(B_RED, 51, 48);
#else
		DrawLite(B_RED, 51, 56);
#endif
	} else if(swapin_lite) {
		swapin_lite = 0;
#ifdef HI_INTS
		DrawLite(B_OFF, 51, 48);
#else
		DrawLite(B_OFF, 51, 56);
#endif
	}

	if (swapouts > last_swapouts && !swapout_lite) {
		swapout_lite = 1;
#ifdef HI_INTS
		DrawLite(B_RED, 56, 48);
#else
		DrawLite(B_RED, 56, 56);
#endif
	} else if(swapout_lite) {
		swapout_lite = 0;
#ifdef HI_INTS
		DrawLite(B_OFF, 56, 48);
#else
		DrawLite(B_OFF, 56, 56);
#endif
	}

	last_swapins = swapins;
	last_swapouts = swapouts;

	first = 0;
}


inline void DrawMem(void)
{
	static int	last_mem = 0, last_swap = 0, first = 1;
	static char buf[128];
	static long mem_total=0;
	static long mem_used=0;
	static long mem_free=0;
	static long mem_shared=0;
	static long mem_buffers=0;
	static long mem_cache=0;
	static long swap_total=0;
	static long swap_used=0;
	static long swap_free=0;

	static int mempercent=0;
	static int swappercent=0;

	memfp = freopen("/proc/meminfo", "r", memfp);

	if (!memfp)
		return;

	while (fgets(buf, 127, memfp)) {
		if (strstr(buf, "Mem:")) {
			sscanf(buf, "Mem: %ld %ld %ld %ld %ld %ld", &mem_total, &mem_used, &mem_free, &mem_shared, &mem_buffers, &mem_cache);
		}

		if (strstr(buf, "Swap:")) {
			sscanf(buf, "Swap: %ld %ld %ld", &swap_total, &swap_used, &swap_free);
		}
	}
            
	mempercent = ((float) (mem_used - mem_buffers - mem_cache) / (float) mem_total) * 100;

	swappercent = ((float) (swap_used) / (float) swap_total) * 100;

/*	printf("mempercent: %i\n", mempercent);
	printf("swappercent: %i\n", swappercent);
*/
    

	if(mempercent != last_mem || first) {
		last_mem = mempercent;
#ifdef HI_INTS
		DrawBar(67, 36, 58,4 , mempercent, 3, 4);
#else
		DrawBar(67, 36, 58, 6, mempercent, 3, 4);
#endif
	}
	    
	if(swappercent != last_swap || first) {
		last_swap = swappercent;
#ifdef HI_INTS
		DrawBar(67, 36, 58, 4, swappercent, 3, 12);
#else
		DrawBar(67, 36, 58, 6, swappercent, 3, 15);
#endif
	}

	first = 0;
}


/* Blits a string at given co-ordinates */
void BlitString(char *name, int x, int y)
{
	static int	i, c, k;

#ifdef MONDEBUG
	puts("BlitString()");
#endif

	k = x;
	for (i=0; name[i]; i++) {

		c = toupper(name[i]); 
		if (c >= 'A' && c <= 'Z') {   // its a letter
			c -= 'A';
			copyXPMArea(c * 6, 74, 6, 8, k, y);
			k += 6;
		} else {   // its a number or symbol
			c -= '0';
			copyXPMArea(c * 6, 64, 6, 8, k, y);
			k += 6;
		}
	}
}


void BlitNum(int num, int x, int y)
{
	static char buf[1024];
	static int newx;

	newx = x;

	sprintf(buf, "%03i", num);

	BlitString(buf, newx, y);
}
    

/* ReadConfigSetting */
int ReadConfigString(FILE *fp, char *setting, char *value)
{
	char str[1024];
	char buf[1024];
	int i;
	int len;
	int slen;
	char *p=NULL;


	if (!fp) return 0;

	sprintf(str, "%s=", setting);
	slen = strlen(str);
    
	fseek(fp, 0, SEEK_SET);

	while ( !feof(fp) ) {
        
		if (!fgets(buf, 512, fp)) break;
        
		len = strlen(buf);

		/* strip linefeed */
		for (i=0; i!=len; i++) {
			if (buf[i] == '\n') {
				buf[i] = 0;
			}
		}

        /* printf("Scanning '%s'...\n", buf); */
		if ( strncmp(buf, str, strlen(str)) == 0) {
            /* found our setting */
            
			for(i=0; i!=slen; i++) {
				if ( buf[i] == '=' ) {
					p=buf+i+1;
					strcpy(value, p);
					return 1;
				}
			}
    
		}
	}
    
        return 0;
}


int ReadConfigInt(FILE *fp, char *setting, int *value)
{
	char buf[1024];

	if (ReadConfigString(fp, setting, (char *) &buf)) {
		*value = atoi(buf);
		return 1;
	}

	return 0;
}


int Read_Config_File( char *filename )
{
	FILE *fp;

	fp = fopen(filename, "r");
	if (fp) {
        
		fclose(fp);
		return 1;
	} else {
		perror("Read_Config_File");
		fprintf(stderr, "Unable to open %s, no settings read.\n", filename);
		return 0;
	}
}


void usage(void)
{
	fprintf(stderr, "\nwmsysmon - http://www.gnugeneration.com\n");
	fprintf(stderr, "\n-------------------\n"
			"|[---------------]|  <--- Memory Use %\n"
			"|[---------------]|  <--- Swap Use %\n"
			"|[---------------]|  <--- I/O %\n"
			"|                 |\n"
			"|   000:00:00     |  <--- Uptime days:hours:minutes\n"
			"|                 |\n"
#ifdef HI_INTS
			"| 01234567   UV   |  <--- 0-N are hardware interrupts 0-24\n"
			"| 89ABCDEF   WX   |  <--- U,V are Page IN/OUT, W,X are Swap IN/OUT\n"
			"| GHIJKLMN   YZ   |\n"
#else
			"| 01234567   WX   |  <--- 0-F are hardware interrupts 0-15\n"
			"| 89ABCDEF   YZ   |  <--- W,X are Page IN/OUT, W,X are Swap IN/OUT\n"
#endif
			"-------------------\n");

	fprintf(stderr, "usage:\n");
	fprintf(stderr, "\t-display <display name>\n");
	fprintf(stderr, "\t-geometry +XPOS+YPOS\tinitial window position\n");
	fprintf(stderr, "\t-r\t\t\tupdate rate in milliseconds (default:300)\n");
	fprintf(stderr, "\t-h\t\t\tthis help screen\n");
	fprintf(stderr, "\t-v\t\t\tprint the version number\n");
	fprintf(stderr, "\n");
}


void printversion(void)
{
	fprintf(stderr, "wmsysmon v%s\n", WMSYSMON_VERSION);
}
